/*
 * Hack by Shin'ichi Konomi June 11, 2005
 * 
 * Based on:
 * 
 * Copyright 1999-2004 Carnegie Mellon University.
 * Portions Copyright 2004 Sun Microsystems, Inc.
 * Portions Copyright 2004 Mitsubishi Electric Research Laboratories.
 * All Rights Reserved.  Use is subject to license terms.
 *
 * See the file "license.terms" for information on usage and
 * redistribution of this file, and for a DISCLAIMER OF ALL
 * WARRANTIES.
 *
 */

package edu.colorado.cs.l3d.clever.affect;

import java.beans.PropertyVetoException;
import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.HttpURLConnection;
import java.net.ProtocolException;
import java.net.URL;
import java.util.Locale;

import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.DataLine;
import javax.sound.sampled.SourceDataLine;
import javax.speech.AudioException;
import javax.speech.Central;
import javax.speech.EngineException;
import javax.speech.EngineList;
import javax.speech.EngineStateError;
import javax.speech.synthesis.Synthesizer;
import javax.speech.synthesis.SynthesizerModeDesc;
import javax.speech.synthesis.Voice;

import com.Ostermiller.util.Browser;

import edu.cmu.sphinx.frontend.util.Microphone;
import edu.cmu.sphinx.recognizer.Recognizer;
import edu.cmu.sphinx.result.Result;
import edu.cmu.sphinx.util.props.ConfigurationManager;


public class AffectiveSpeech {
//	 Used to identify the windows platform.
    private static final String WIN_ID = "windows";
    // The default system browser under windows.
    private static final String WIN_PATH = "rundll32";
    // The flag to display a url.
    private static final String WIN_FLAG = "url.dll,FileProtocolHandler";
    // The default browser under unix.
    private static final String UNIX_PATH = "netscape";
    // The flag to display a url.
    private static final String UNIX_FLAG = "-remote openURL";
    //
    private static final String MAC_ID = "mac";
    // 

    
	private static final int EXTERNAL_BUFFER_SIZE = 128000;
	
    /**
     * Returns a "no synthesizer" message, and asks 
     * the user to check if the "speech.properties" file is
     * at <code>user.home</code> or <code>java.home/lib</code>.
     *
     * @return a no synthesizer message
     */
    static private String noSynthesizerMessage() {
        String message =
            "No synthesizer created.  This may be the result of any\n" +
            "number of problems.  It's typically due to a missing\n" +
            "\"speech.properties\" file that should be at either of\n" +
            "these locations: \n\n";
        message += "user.home    : " + System.getProperty("user.home") + "\n";
        message += "java.home/lib: " + System.getProperty("java.home") +
	    File.separator + "lib\n\n" +
            "Another cause of this problem might be corrupt or missing\n" +
            "voice jar files in the freetts lib directory.  This problem\n" +
            "also sometimes arises when the freetts.jar file is corrupt\n" +
            "or missing.  Sorry about that.  Please check for these\n" +
            "various conditions and then try again.\n";
        return message;
    }

    /**
     * Example of how to list all the known voices for a specific
     * mode using just JSAPI.  FreeTTS maps the domain name to the
     * JSAPI mode name.  The currently supported domains are
     * "general," which means general purpose synthesis for tasks
     * such as reading e-mail, and "time" which means a domain that's
     * only good for speaking the time of day. 
     */
    public static void listAllVoices(String modeName) {
        
        System.out.println();
        System.out.println(
            "All " + modeName + " Mode JSAPI Synthesizers and Voices:");

        /* Create a template that tells JSAPI what kind of speech
         * synthesizer we are interested in.  In this case, we're
         * just looking for a general domain synthesizer for US
         * English.
         */ 
        SynthesizerModeDesc required = new SynthesizerModeDesc(
            null,      // engine name
            modeName,  // mode name
            Locale.US, // locale
            null,      // running
            null);     // voices

        /* Contact the primary entry point for JSAPI, which is
         * the Central class, to discover what synthesizers are
         * available that match the template we defined above.
         */
        EngineList engineList = Central.availableSynthesizers(required);
        for (int i = 0; i < engineList.size(); i++) {
            
            SynthesizerModeDesc desc = (SynthesizerModeDesc) engineList.get(i);
            System.out.println("    " + desc.getEngineName()
                               + " (mode=" + desc.getModeName()
                               + ", locale=" + desc.getLocale() + "):");
            Voice[] voices = desc.getVoices();
            for (int j = 0; j < voices.length; j++) {
                System.out.println("        " + voices[j].getName());
            }
        }
    }
    

    public static void main(String[] args) throws IOException {    	
    	if (args.length > 0 ){
          	String baseURL = args[0];
          	String[] data;
          	
          	if (args.length > 1){
          		////////////////////////////////////////////
          		//data = listAlbums(baseURL);
          		//data = getClient(baseURL, "MicrosoftWorld");
          		//data = getQuestion(baseURL, "MicrosoftWorld");
          		//data = countPhotos(baseURL, "MicrosoftWorld");
          		//for(int i = 0; i< data.length; i++){
          		//	System.out.println(data[i]);
          		//}
          		///////////////////////////////////////////
          		String album = args[1];
          		data = getClient(baseURL, album);
          		String clientName = "Nobody";
          		if (data != null && data.length > 0) clientName = data[0];
          		data = getQuestion(baseURL, album);
          		String question = "Do you like this picture?";
          		if (data != null && data.length > 0) question = data[0];
          		data = countPhotos(baseURL, album);
          		int numPhotos = 0;
          		if (data != null && data.length > 0) numPhotos = Integer.parseInt(data[0]);
        		Browser.init();
          		///////////////////////////////////////////
    	        URL configURL = AffectiveSpeech.class.getResource("config.xml");
    	        System.out.println("Setting up Synthesizer...");
    	        String voiceName = selectVoice(null);
    	        try {
    	         	Synthesizer synthesizer = setUpSynthesizer(voiceName);
    	        	System.out.println("Loading...");
    	     		playSound("single.wav"); 
    				synthesizer.speakPlainText("Hello, " + clientName, null);
    				synthesizer.waitEngineState(Synthesizer.QUEUE_EMPTY);
    				
    	    		//ArrayList photoInfo = new ArrayList();
    	    		//setPhotoInfo(photoNamePrefix, photoNamePostfix, numPhotos, defaultRating, photoInfo);
    				//showPhoto(photoInfo,0);
    				
    	        	ConfigurationManager cm = new ConfigurationManager(configURL);
    	        	Recognizer recognizer = (Recognizer) cm.lookup("recognizer");
    	        	Microphone microphone = (Microphone) cm.lookup("microphone");
    	        	/* allocate the resource necessary for the recognizer */
    	        	recognizer.allocate();
    	        	/* the microphone will keep recording until the program exits */
    	        	for (int i = 0; i < numPhotos; i++){
    	         		String s1 = "";
    	         		Result result = null;
    					showPhoto(baseURL, album, i);
    					if(i % 2 == 0){
    						synthesizer.speakPlainText("So, " + clientName + ", " + 
    								question, null);
    						synthesizer.waitEngineState(Synthesizer.QUEUE_EMPTY);
    					} else {
    						synthesizer.speakPlainText("How about this one? " +
    								question, null);
    						synthesizer.waitEngineState(Synthesizer.QUEUE_EMPTY);	    					
    					}
    					System.out.println("Start speaking. Press Ctrl-C to quit.\n");
    					/*
    					 * This method will return when the end of speech
    					 * is reached. Note that the endpointer will determine
    					 * the end of speech.
    					 */ 
    					microphone.clear();
    	             	if (microphone.startRecording()) {
    	    				result = recognizer.recognize();
    	    				microphone.stopRecording();
    		    		} else {
    		    			System.out.println("Cannot start microphone.");
    		    			recognizer.deallocate();
    		    			System.exit(1);
    		    		}
    					if (result != null) {
    						String resultText = result.getBestFinalResultNoFiller();
    						s1 = "You said: "+resultText;
    						System.out.println(s1 + "\n");
    						/* The the synthesizer to speak and wait for it to
    						 * complete.
    						 */
    						synthesizer.speakPlainText(s1, null);
    						synthesizer.waitEngineState(Synthesizer.QUEUE_EMPTY);
    						updateScore(resultText, baseURL, album, i);
    	 				} else {
    						System.out.println("I can't hear what you said.\n");
    					}
    				}
    				synthesizer.speakPlainText("Thank you, " + clientName + ".", null);
    	    		showNewsLetter(baseURL, album);
    				synthesizer.speakPlainText("Here's your newsletter. Yo can print it if you like.", null);
    				synthesizer.waitEngineState(Synthesizer.QUEUE_EMPTY);
    				finishSynthesizer(synthesizer);
    	        } catch (Exception e) {
    	        	e.printStackTrace();
    	        }          	         		
          	} else {
          		System.out.println("\nYou must specify an album. ");
          		System.out.println("Currently available albums:");
          		data = listAlbums(baseURL);
          		for(int i = 0; i< data.length; i++){
          			System.out.print("    \"");
                  	System.out.print(data[i]);
                  	System.out.println("\"");
                }
          	}
    } else { // no args
  		System.out.println("\nserverURL not specified\n");
  		System.out.println("Usage: start.bat serverURL albumNam");
    	System.out.println("e.g., start.bat http://l3d.cs.colorado.edu/~konomi/demo/albums/index.php\n");
    }
}

	/**
	 * @param baseURL
	 */
	private static String[] listAlbums(String baseURL) {
		String[] result =null;
		try {
			URL serverURL = new URL(baseURL + "?command=listAlbums");
			String data = httpGet(serverURL);
			result = data.split("<br>");
		} catch (Exception e){
			e.printStackTrace();
			System.exit(0);
		}
		return result;
	}
	
	/**
	 * @param baseURL
	 */
	private static String[] getClient(String baseURL, String album) {
		String[] result =null;
		try {
			URL serverURL = new URL(baseURL + "?command=getClient&directory=" + album);
			String data = httpGet(serverURL);
			result = data.split("<br>");
		} catch (Exception e){
			e.printStackTrace();
			System.exit(0);
		}
		return result;
	}
	/**
	 * @param baseURL
	 */
	private static String[] getQuestion(String baseURL, String album) {
		String[] result =null;
		try {
			URL serverURL = new URL(baseURL + "?command=getQuestion&directory=" + album);
			String data = httpGet(serverURL);
			result = data.split("<br>");
		} catch (Exception e){
			e.printStackTrace();
			System.exit(0);
		}
		return result;
	}
	/**
	 * @param baseURL
	 */
	private static String[] countPhotos(String baseURL, String album) {
		String[] result =null;
		try {
			URL serverURL = new URL(baseURL + "?command=countPhotos&directory=" + album);
			String data = httpGet(serverURL);
			result = data.split("<br>");
		} catch (Exception e){
			e.printStackTrace();
			System.exit(0);
		}
		return result;
	}
	
	/**
	 * @param serverURL
	 * @throws IOException
	 * @throws ProtocolException
	 */
	public static String httpGet(URL serverURL) throws IOException, ProtocolException {
		HttpURLConnection urlConn;
		urlConn = (HttpURLConnection) serverURL.openConnection();
		urlConn.setRequestMethod("GET");
		BufferedReader br = new BufferedReader(
				new InputStreamReader(urlConn.getInputStream()));
		StringBuffer sb = new StringBuffer();	
		while(true){
			String line = br.readLine();
			if(line == null)break;
			sb.append(line);
		}
		br.close();
		urlConn.disconnect();
		return sb.toString();
	}

	/**
	 * @param numPhotos
	 * @param photoInfo
	 */
	private static void showNewsLetter(String baseURL, String album) {
		showPage(baseURL, "newsletter", album, null, -1);
	}

	/**
	 * @param resultText
	 * @param photoInfo
	 * @param i
	 */
	private static void updateScore(String resultText, 
			String baseURL, 
			String album, 
			int i) {
		if (resultText.equals("yes") || 
			resultText.equals("yeah") || 
			resultText.equals("alright")){
			System.out.println("Photo score up");
			// use the number for the next picture (i.e., i+1)
			showPage(baseURL, "newsletter", album, "yes", i+1);
		} else {
			System.out.println("Photo score down");
			// use the number for the next picture (i.e, i+1)
			showPage(baseURL, "newsletter", album, "no", i+1);
		}
	}

	/**
	 * @param i
	 * @param photoInfo
	 */
	private static void showPhoto(String baseURL, String album, int i) {
		showPage(baseURL, "newsletter", album, null, i);
	}
	
	/**
	 * @param i
	 * @param photoInfo
	 * @throws IOException
	 */
	private static void showPage(String baseURL, 
			String mode, 
			String album, 
			String ans, 
			int i) {
		String showURL = baseURL + (mode != null ? "?mode=" + mode : "")  
			+ (album != null ? "&directory=" + album : "") 
			+ (ans != null ? "&ans=" + ans : "") 
			+ (i >= 0 ? "&image=" +i : "");
		try {
			Browser.displayURL(showURL);
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		//if (osPlatform().equals("windows")){
		//execute(new String[]{WIN_PATH,
		//		WIN_FLAG, 
		//		showURL}); 
		//} else {
		//	System.out.println("Non Windows Platform.");
		//}
	}
	
	/**
	 * @param photoNamePrefix
	 * @param photoNamePostfix
	 * @param numPhotos
	 * @param defaultScore
	 * @param photoInfo
	 */
/*
	private static void setPhotoInfo(String photoNamePrefix, String photoNamePostfix, int numPhotos, short defaultScore, ArrayList photoInfo) {
		for (int i = 0; i < numPhotos; i++) {
			StringBuffer sb = new StringBuffer();
		    sb.append(photoNamePrefix);
			sb.append(Integer.valueOf(i+1).toString());
			sb.append(".");
			sb.append(photoNamePostfix);
			photoInfo.add(sb.toString());
			photoInfo.add(Integer.valueOf(defaultScore));						
		}
	}
	*/

	/**
	 * @param args
	 * @return
	 */

	private static String selectVoice(String[] args) {
		/* List all the "general" domain voices, which are voices that
         * are capable of attempting to speak almost any text you
         * throw at them.
         */
		String voiceName;
        listAllVoices("general");

        if (args != null){
        	voiceName = (args.length > 0) ? args[0] : "kevin16";
        } else {
        	voiceName = "kevin16";
        }
        System.out.println();
        System.out.println("Using voice: " + voiceName);
		return voiceName;
	}

	/**
	 * @param synthesizer
	 * @throws EngineException
	 * @throws EngineStateError
	 */
		
	private static void finishSynthesizer(Synthesizer synthesizer) throws EngineException, EngineStateError {
		/* Clean up and leave.
             */
	    synthesizer.deallocate();
            System.exit(0);
	}

	/**
	 * @param voiceName
	 * @return
	 * @throws EngineException
	 * @throws EngineStateError
	 * @throws AudioException
	 * @throws PropertyVetoException
	 */
	private static Synthesizer setUpSynthesizer(String voiceName) throws EngineException, EngineStateError, AudioException, PropertyVetoException {
		/* Find a synthesizer that has the general domain voice
             * we are looking for.  NOTE:  this uses the Central class
             * of JSAPI to find a Synthesizer.  The Central class
             * expects to find a speech.properties file in user.home
             * or java.home/lib.
             *
             * If your situation doesn't allow you to set up a
             * speech.properties file, you can circumvent the Central
             * class and do a very non-JSAPI thing by talking to
             * FreeTTSEngineCentral directly.  See the WebStartClock
             * demo for an example of how to do this.
             */
	    SynthesizerModeDesc desc = new SynthesizerModeDesc(
                null,          // engine name
                "general",     // mode name
                Locale.US,     // locale
                null,          // running
                null);         // voice
	    Synthesizer synthesizer = Central.createSynthesizer(desc);

            /* Just an informational message to guide users that didn't
             * set up their speech.properties file. 
             */
	    if (synthesizer == null) {
		System.err.println(noSynthesizerMessage());
		System.exit(1);
	    }

	    /* Get the synthesizer ready to speak
             */
	    synthesizer.allocate();
	    synthesizer.resume();

            /* Choose the voice.
             */
            desc = (SynthesizerModeDesc) synthesizer.getEngineModeDesc();
            Voice[] voices = desc.getVoices();
            Voice voice = null;
            for (int i = 0; i < voices.length; i++) {
                if (voices[i].getName().equals(voiceName)) {
                    voice = voices[i];
                    break;
                }
            }
            if (voice == null) {
                System.err.println(
                    "Synthesizer does not have a voice named "
                    + voiceName + ".");
                System.exit(1);
            }
            synthesizer.getSynthesizerProperties().setVoice(voice);
		return synthesizer;
	}
	
	/**
	 * Play a sound file
	 * 
	 * @param fname
	 */
   private static void playSound(String fname){
        try {
	          File soundFile = new File(fname);
	          AudioInputStream audioInputStream = AudioSystem.getAudioInputStream(soundFile);
	          AudioFormat audioFormat = audioInputStream.getFormat();
	
	          DataLine.Info info = new DataLine.Info(SourceDataLine.class,audioFormat);
	          SourceDataLine line = (SourceDataLine) AudioSystem.getLine(info);
	          line.open(audioFormat);
	          line.start();
	
	          int nBytesRead = 0;
	          byte[] abData = new byte[EXTERNAL_BUFFER_SIZE];
	          while (nBytesRead != -1) {
	            nBytesRead = audioInputStream.read(abData, 0, abData.length);
	            if (nBytesRead >= 0) {
	              int nBytesWritten = line.write(abData, 0, nBytesRead);
	            }
	          }
	          line.drain();
	          line.close();
        } catch (Exception e) {
	          e.printStackTrace();
	          System.exit(1);
        }
    }
    
    /**
     * execute the given command
     *
     * @param cmd the command to execute
     */
    private static void execute(String[] cmd) {
        try {
            Runtime.getRuntime().exec(cmd);
        } catch (IOException e) {
	          e.printStackTrace();
           // if we can't run the command, just fall back to
           // a non-working demo.
        }
    }
    
    /**
     * Try to determine whether this application is running under Windows
     * or some other platform by examing the "os.name" property.
     *
     * @return true if this application is running under a Windows OS
     */
    public static String osPlatform()
    {
    	String result = "";
        String os = System.getProperty("os.name").toLowerCase();
        System.out.print(os);
        System.out.println("; Java Version: " + System.getProperty("java.version"));
        if (os != null) {
        	if ( os.startsWith(WIN_ID))
        		result = "windows";
        	else if ( os.startsWith(MAC_ID))
        		result = "mac";
        	else
        		result = "other";
        }
        return result;
    }

}
